
/*
  Copyright 1995 The University of Rhode Island and The Massachusetts
  Institute of Technology

  Portions of this software were developed by the Graduate School of
  Oceanography (GSO) at the University of Rhode Island (URI) in collaboration
  with The Massachusetts Institute of Technology (MIT).

  Access and use of this software shall impose the following obligations and
  understandings on the user. The user is granted the right, without any fee
  or cost, to use, copy, modify, alter, enhance and distribute this software,
  and any derivative works thereof, and its supporting documentation for any
  purpose whatsoever, provided that this entire notice appears in all copies
  of the software, derivative works and supporting documentation.  Further,
  the user agrees to credit URI/MIT in any publications that result from the
  use of this software or in any product that includes this software. The
  names URI, MIT and/or GSO, however, may not be used in any advertising or
  publicity to endorse or promote any products or commercial entity unless
  specific written permission is obtained from URI/MIT. The user also
  understands that URI/MIT is not obligated to provide the user with any
  support, consulting, training or assistance of any kind with regard to the
  use, operation and performance of this software nor to provide the user
  with any updates, revisions, new versions or "bug fixes".

  THIS SOFTWARE IS PROVIDED BY URI/MIT "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
  EVENT SHALL URI/MIT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
  DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE ACCESS, USE OR PERFORMANCE
  OF THIS SOFTWARE.
*/

// Convert to CDF server
// See README file for initial mods
// csf 4/19/96

// This file contains functions which read the variables and their attributes
// from a netcdf file and build the in-memeory DAS. These functions form the
// core of the server-side software necessary to extract the DAS from a
// netcdf data file.
//
// It also contains test code which will print the in-memeory DAS to
// stdout. It uses both the DAS class as well as the netcdf library.
// In addition, parts of these functions were taken from the netcdf program
// ncdump, from the netcdf standard distribution (ver 2.3.2)
//
// jhrg 9/23/94

// $Log: cdfdas.cc,v $
// Revision 1.1  1996/06/10 20:34:15  jimg
// Added.
//
// Revision 1.15.2.1  1995/10/12  17:10:23  jimg
// Added {}'s to case labels which include variable definitions from which
// values are returned.
//
// Revision 1.15  1995/07/09  21:34:01  jimg
// Added copyright notice.
//
// Revision 1.14  1995/06/29  20:29:50  jimg
// Added delete [] of the return value from print_attr(). This patched a
// reported memory leak (reported from dbnew).
//
// Revision 1.13  1995/06/28  20:22:42  jimg
// Replaced malloc calls with calls to new (and calls to free with calls to
// delete).
//
// Revision 1.12  1995/06/27  20:48:52  jimg
// Changed the scope of print_type() and print_attr() to static.
// Fixed an off-by-one error in read_attribtes() when working with NC_CHAR
// attributes (which we represent as strings).
//
// Revision 1.11  1995/06/23  15:35:23  jimg
// Added netio.h.
// Fixed some misc problems which may have come about do to the modifications
// in nc_das.cc and nc_dds.cc (they went from CGIs to simple UNIX filters).
//
// Revision 1.10  1995/03/16  16:38:35  reza
// Updated byte transfer. Bytes are no longer transmitted in binary but in
// an ASCII string (0-255).
//
// Revision 1.9  1995/02/10  04:47:37  reza
// Updated to use type subclasses, NCArray, NCByte, NCInt32, NCGrid....
//
// Revision 1.8  1994/12/22  04:46:20  reza
// Updated to use DODS new attribute array capability.
//
// Revision 1.7  1994/11/03  05:18:27  reza
// Modified the redundant trailing type-code and illegal (in das parser)
// single quotation marks used for byte type.
//
// Revision 1.6  1994/10/28  15:14:15  reza
// Changed some comments
//
// Revision 1.5  1994/10/06  17:53:07  jimg
// Some reformatting of code for emacs, ...
// Fixed Makefile.in so that two targets are built: client and server.
// Fixed cast away of const in ncdas.cc.
//
// Revision 1.4  1994/10/06  16:29:54  jimg
// Fixed printed representation of Strings read from a netcdf file -- they
// now are correctly enclosed in double quotes.
//
// Revision 1.3  1994/10/06  15:52:10  reza
// Changed error messeges
//
// Revision 1.2  1994/10/05  18:00:17  jimg
// Modified so that types (as represented by the DAS) are now handled
// correctly.
//
// Revision 1.1  1994/09/27  23:19:37  jimg
// First version of the netcdf software which used libdas++.a.
//

static char rcsid[]={"$Id: cdfdas.cc,v 1.1 1996/06/10 20:34:15 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <iostream.h>

#include "netio.h"
#include "cdf.h"
#include "cdfUtil.h"
#include "DAS.h"


// These are used in to get as the return values for print_type().

const char STRING[]="String";
const char BYTE[]="Byte";
const char INT32[]="Int32";
const char FLOAT64[]="Float64";

// Used by ErrMsgT
char Msgt[255];


/****************************************************************************
  This routine returns the length in bytes of the standard CDF types
****************************************************************************/
const char *
cdfToDodsType(long cdfType)
{
  switch (cdfType) {
  case CDF_BYTE:	
	return BYTE;

  case CDF_CHAR:	
  case CDF_UCHAR:	
	return STRING;

  case CDF_INT1:
  case CDF_UINT1:
  case CDF_INT2:
  case CDF_UINT2:	
  case CDF_INT4:
  case CDF_UINT4:	
	return INT32;

  case CDF_REAL4:	
  case CDF_FLOAT:	
  case CDF_REAL8:	
  case CDF_DOUBLE:
  case CDF_EPOCH:	
	return FLOAT64;

  }
  return STRING;
}
/****************************************************************************
  This routine returns the value of the attribute as an ascii string
****************************************************************************/
char *
print_attr(long attrType, void * value)
{
  char *ascRep;	
  char c = 'C';
  unsigned char uc;
  int s;
  unsigned int us;
  long l;
  unsigned long ul;
  float f;
  double d;
  char es[EPOCH_STRING_LEN+1];
 
 
  // Allocate storage for strings or other types
  if (attrType == CDF_CHAR || attrType == CDF_UCHAR) 
	ascRep = new char [strlen((const char *)value) + 3];
  else
	ascRep = new char [32];
  
  switch (attrType) {
  case CDF_CHAR:	
  case CDF_UCHAR: 
	sprintf(ascRep, "\"%s\"", (char*)value);
	break;

  case CDF_INT1: 
	c = *(char *)value;
	sprintf(ascRep, "%d", c);
	break;

  case CDF_BYTE:	
  case CDF_UINT1: 
	c = *(unsigned char *)value;
	sprintf(ascRep, "%d", uc);
	break;

  case CDF_INT2: 
	s = *(int *)value;
	sprintf(ascRep, "%d", s);
	break;

  case CDF_UINT2:
	us = *(unsigned int *)value;
	sprintf(ascRep, "%d", us);
	break;

  case CDF_INT4: 
	l = *(long *)value;
	sprintf(ascRep, "%d", l);
	break;

  case CDF_UINT4:
	ul = *(unsigned long *)value;
	sprintf(ascRep, "%d", ul);
	break;

  case CDF_REAL4:	
  case CDF_FLOAT:
	f = *(float *)value;
	sprintf(ascRep, "%g", f);
	if (!strchr(ascRep, '.') && !strchr(ascRep, 'e'))
	  strcat(ascRep, ".");
	break;

  case CDF_REAL8:	
  case CDF_DOUBLE:
  case CDF_EPOCH: 
	d = *(double *)value;
	sprintf(ascRep, "%g", d);
	if (!strchr(ascRep, '.') && !strchr(ascRep, 'e'))
	  strcat(ascRep, ".");
	break;

	/**
	d = *(double *)value;
	encodeEPOCH(d, es);
	sprintf(ascRep, "\"%s\"", es);
	break;
	**/

  default: 
	sprintf (Msgt, "cdf_das: Invalid CDF Type: %dl", attrType);
	ErrMsgT(Msgt);
	break;
  }

  return ascRep;
}

/****************************************************************************
  This reads and processes an attribute entry
****************************************************************************/
bool
read_attr_entries(DAS &das, CDFid cdfid, char* fileName, long attrId, 
				  char* attrName, long attrScope, long attrEntryId)
{
  CDFstatus status;
  long attrType, saveType;
  bool initSave = false;
  long numAttrElts;
  long varType;
  long numVarElts;
  long recVar;
  long dimVar[CDF_MAX_DIMS];
  char varName[CDF_VAR_NAME_LEN+1];
  AttrTable *attr_table_ptr;
  int numElts;
  
  // Retrieve attribute entry info
  status = CDFattrEntryInquire(cdfid, attrId, attrEntryId,
							   &attrType, &numAttrElts);
  // Continue only if entry found
  if (status != NO_SUCH_ENTRY) {
	if (status != CDF_OK) {
	  sprintf (Msgt, "cdf_das: Invalid attribute entry number. " 
			   "File: %s Attribute: %d Entry: %d", 
			   fileName, attrId, attrEntryId);
	  ErrMsgT(Msgt);
	  return false;
	}
	
	// DODS wants all values of an attributes to have the same type. 
	// Currently returns if multiple types, however it might be 
	// better just to skip the entry.
	if (!initSave) {
	  saveType = attrType;
	  initSave = true;
	}
	else 
	  if (attrType != saveType) {
		sprintf (Msgt, "cdf_das: Multiple datatypes for attribute. " 
				 "File: %s Attribute: %d Entry: %d", 
				 fileName, attrId, attrEntryId);
		ErrMsgT(Msgt);
		return false;
	  }
	
	// Set variable name
	if (attrScope == GLOBAL_SCOPE) {
	  strcpy(varName, "NC_GLOBAL");
	}
	else {
	  status = CDFvarInquire(cdfid, attrEntryId, 
							 varName, &varType, &numVarElts, &recVar, dimVar);
	  if (status != CDF_OK) {
		sprintf (Msgt, "cdf_das: Invalid variable number. " 
				 "File: %s Attribute: %d Variable: %d", 
				 fileName, attrId, attrEntryId);
		ErrMsgT(Msgt);
		return false;
	  }
	}

	// Retrieve or create the attribute table for each variable
	attr_table_ptr = das.get_table(varName);
	if (!attr_table_ptr)
	  attr_table_ptr = das.add_table(varName, new AttrTable);
	
	// Allocate space and get the attribute value
	void * value = (void *) new char [(numAttrElts+1) * cdfTypeLen(attrType)];
	if (!value) {
	  ErrMsgT("cdf_das: Out of memory!");
	  return false;
	}
	status = CDFattrGet(cdfid, attrId, attrEntryId, value);
	if (status != CDF_OK) {
	  sprintf (Msgt, "cdf_das: Could not read attribute value "
			   "File: %s Attribute: %d", fileName, attrId);
	  ErrMsgT(Msgt);
	  return false;
	}

	// If the datatype is character then we have a string. CDF
	// represents strings as arrays of char, but we represent them as X
	// strings. So... Add the null and set the number of elements to 1
	if (attrType == CDF_CHAR || attrType == CDF_UCHAR) { 
	  *((char *)value + numAttrElts) = '\0';
	  numElts = 1;
	} 
	else {
	  numElts = numAttrElts;
	}
	
	// add all the attributes in the array to dods attribute table
	// Loop through all element, calculate offset and get ascii representation
	for (int i=0; i < numElts ; i++) {
	  int loc = i * cdfTypeLen(attrType);
	  char * print_rep = print_attr(attrType, value+loc);
	  //  cout << varName << " " << attrName  << " " << print_rep << endl;
	  attr_table_ptr->append_attr(attrName, cdfToDodsType(attrType), 
								  print_rep);
	  delete [] print_rep;
	}
	
	// Free storage for attribute value
	delete [] value;
  }
  return true;                        
}

/****************************************************************************

 Given a reference to an instance of class DAS and a filename that
 refers to a cdf file, read the file and extract all the global and
 variable attribute entries.  Add the Add the attributes to the
 instance of DAS.  Returns: false if an error accessing the cdf file
 was detected, true otherwise.

****************************************************************************/
bool
read_attributes(DAS &das, char *fileName)
{
    CDFid cdfid;
	CDFstatus status;
    long ndims, nvars, natts, encoding, majority, maxRec;
	long dimSizes[CDF_MAX_DIMS];
	char attrName[CDF_VAR_NAME_LEN+1];
	long attrScope;
	long maxEntry;

	status = CDFopen(fileName, &cdfid);
    if (status != CDF_OK) {
        sprintf (Msgt,"cdf_das: Could not open file %s",fileName);
        ErrMsgT(Msgt);
		return false;
    }

    // how many variables? how many global attributes? 
    if (CDFinquire(cdfid, &ndims, dimSizes, &encoding, &majority, 
				   &maxRec, &nvars, &natts) != CDF_OK) {
	  ErrMsgT("cdf_das: Could not inquires about CDF file");
	  return false;
    }

	// Retrieve all attributes
	for (long attrId = 0; attrId<natts; attrId++){

	  // Retrieve attribute info
	  status = CDFattrInquire(cdfid, attrId, attrName, &attrScope, &maxEntry);
	  if (status != CDF_OK) {
        sprintf (Msgt, "cdf_das: Invalid attribute number. " 
				 "File: %s Attribute: %d", fileName, attrId);
        ErrMsgT(Msgt);
		return false;
	  }

	  // Retrieve attribute entries
	  for (long entry = 0; entry<=maxEntry; entry++) 
		read_attr_entries(das, cdfid, fileName, 
						  attrId, attrName, attrScope, entry);
	}

	CDFclose(cdfid);
    return true;
}

/**********************************************************************
  Routines for testing
**********************************************************************/
int
atts()
{
  DAS das;
  char name[] = "wind.cdf";
  int i = 2;

  char * rep = print_attr(CDF_INT2, (void*)&i);

  bool b = read_attributes(das, name);
  return 0;
}

#ifdef TEST
int
main(int argc, char *argv[])
{
    DAS das;

    if(!read_attributes(das, "wind.cdf"))
	abort();

    das.print();
}


#endif



